/* -----------------------------------------------------------------------------
*
*  nposcpack.h
*
*  ANTz - realtime 3D data visualization tools for the real-world, based on NPE.
*
*  ANTz is hosted at http://openantz.com and NPE at http://neuralphysics.org
*
*  Written in 2010-2013 by Shane Saxon - contact@saxondigital.net
*
*  Please see main.c for a complete list of additional code contributors.
*
*  To the extent possible under law, the author(s) have dedicated all copyright 
*  and related and neighboring rights to this software to the public domain
*  worldwide. This software is distributed without any warranty.
*
*  Released under the CC0 license, which is GPL compatible.
*
*  You should have received a copy of the CC0 Public Domain Dedication along
*  with this software (license file named COPYING.txt). If not, see
*  http://creativecommons.org/publicdomain/zero/1.0/
*
* --------------------------------------------------------------------------- */
/* 
*  This file provides a C-Wrapper for the C++ based oscpack library
*
* --------------------------------------------------------------------------- */

#ifndef NPOSCPACK_H_
#define NPOSCPACK_H_


struct OscPackListener {
	int oscPackIdx;  // index into array that holds the oscpack instances
};
typedef struct OscPackListener OscPackListener;
typedef struct OscPackListener* pOscPackListener;

// listener functions
void npStartListeners( pOscPackListener oscListener, int argc, char** argv, void* dataRef );
void npInitOscPackListener( pOscPackListener oscListener, int port, void* dataRef );
void npStartOscPackListener( pOscPackListener oscListener );

// read a line that has been received by listener
char* npOscReadLine( char* bufferStart, int maxLength, int udpFd );

struct OscPackSender {
	int oscPackIdx; // index into array that holds the oscpack instances
};

typedef struct OscPackSender OscPackSender;
typedef struct OscPackSender* pOscPackSender;

// sender functions
void npInitOscPackSender( pOscPackSender oscSender, char* ip, int port );
void npCreateOscPackBundle( pOscPackSender oscSender );
void npBeginOscPackMessage( pOscPackSender oscSender, char* message );
void npOscAddString( pOscPackSender oscSender, char* s );
void npOscAddInt( pOscPackSender oscSender, int i );
void npOscAddFloat( pOscPackSender oscSender, float f );
void npOscAddBool( pOscPackSender oscSender, bool b );
void npEndOscPackMessage( pOscPackSender oscSender );
void npSendOscPackBundle( pOscPackSender oscSender );

#define OSC_csvCh "/csvCh"
#define OSC_csvChMap "/csvChMap"
#define OSC_csvNode "/csvNode"
#define OSC_csvTag "/csvTag"

#define READLINE "/readline"
#define DUMP "/dump"


#endif

